#include "EEPROM.h"
#include "Wire.h"
#include "ZUNO_OLED_I2C.h"

// Set this flag to enable debug output in the serial monitor
#define DEBUGMODE

#ifdef DEBUGMODE
  #define DEBUG(s) Serial.print(s)
  #define DEBUGLN(s) Serial.println(s)
  #define DEBUG2(s,f) Serial.print(s,f)
  #define DEBUGLN2(s,f) Serial.println(s,f)
  
#else
  #define DEBUG(s)
  #define DEBUGLN(s)
  #define DEBUG2(s,f)
  #define DEBUGLN2(s,f)
#endif

#define PULSE_LENGTH 100

/* Constants for EEPROM access
 * 
 */
#define EEPROM_ADDR             0x800
#define EEPROM_UPDATE_INTERVAL  3600000

/* Energy meter ticks
 *  
 * How many pulses/ticks per 1 kWh 
 * Default setting: 1000
 */
#define TICKS_PER_KWH_PARAMETER 64
#define TICKS_PER_KWH_DEFAULT 1000
unsigned long ppwh;
unsigned long ppkwh;

/* Power reports
 *  
 * Minimum change in consumed power that will result in sending new power report to the main controller.
 * 0 – reports are disabled
 * 1-100 (1-100%) – change in power
 * Default setting: 5 (5%)
 */
#define POWER_CHANGE_REPORT_PARAMETER 65
#define POWER_CHANGE_REPORT_DEFAULT 5
unsigned long p_changereport;

/* Time between power reports
 *  
 * Minimum time that has to elapse before sending new power report to the main controller.
 * 0 – reports are disabled
 * 1-3600 (1-3600s) – report interval
 * Default setting: 10 (10s)
 */
#define POWER_TIME_REPORT_PARAMETER 66
#define POWER_TIME_REPORT_DEFAULT 10
unsigned long p_timereport;

/* Energy reports
 *  
 * The minimum change in consumed energy that will result in sending new energy report to the main controller.
 * 0 – reports are disabled
 * 1-32000 (1 – 32000 kWh) – change in energy
 * Default setting: 1 (1kWh)
 */
#define ENERGY_CHANGE_REPORT_PARAMETER 67
#define ENERGY_CHANGE_REPORT_DEFAULT 1
unsigned long e_changereport;

/* Time between energy reports
 *  
 * Minimum time that has to elapse before sending new energy report to the main controller.
 * 0 – reports are disabled
 * 1-3600 (1-3600s) – report interval
 * Default setting: 60 (60s)
 */
#define ENERGY_TIME_REPORT_PARAMETER 68
#define ENERGY_TIME_REPORT_DEFAULT 15
unsigned long e_timereport;

/* Energy seed
 *  
 * For forcing the z-uno to initialize the energy counter at a specif value
 * After the seed has been received, this paramter will be cleared in the controller
 * 
 * 1-32000 (1-32000 kWh) - seed value
 * Default setting: 0 (0 kWh)
 */
#define ENERGY_SEED_PARAMETER 69
#define ENERGY_SEED_DEFAULT 0
unsigned long e_seed;

float currentpower;
float previouspower;
float powerincrease;
float powerdelta;
unsigned long previouspowerreport;

float currentenergy;
float previousenergy;
float energyincrease;
unsigned long previousenergyreport;

// Timestammps
unsigned long pulsecount = 0;
unsigned long previouspulsecount = 0;
unsigned long pulsestart = 0; // timestamp when the pulse started
unsigned long previouspulse = 0;
unsigned long pulsedistance = 0;
unsigned long currenttime = 0;
unsigned long lastEEPROMsave = 0;

int EEPROMreturnval; // For capturing the return value from EEPROM operations

bool pulseactive = false;

bool sendpowerreport = false;
bool sendenergyreport = false;

// Interrupt pin where the pulse counter is attached
int pulsepin = 3;

// Output LED which will blnk at same frequency as meter signal
s_pin signalpin = 13;

// Driver for the oled display
//OLED oled;

// EEPROM data structure
struct meter_data {
    unsigned long pulsecount;
    byte          crc8;
};

meter_data my_meter_data;

byte crc8(byte * data, byte count) {
    byte result = 0xDF;

    while(count--) {
        result ^= *data;
        data++;
    }
    return result;
}

void save_meter_data() { 
    my_meter_data.crc8 = crc8((byte*)&my_meter_data, sizeof(meter_data) - 1);
    EEPROMreturnval = EEPROM.put(EEPROM_ADDR, &my_meter_data, sizeof(meter_data));
    DEBUG("Saving to EEPROM. code = ");
    DEBUG(EEPROMreturnval);
    DEBUG(", Pulsecount = ");
    DEBUGLN(my_meter_data.pulsecount);
    
}

/* Z-Wave channels
 * 
 */
//ZUNO_SETUP_CHANNELS(
//    ZUNO_METER(
//      ZUNO_METER_TYPE_ELECTRIC, 
//      METER_RESET_ENABLE, 
//      ZUNO_METER_ELECTRIC_SCALE_WATTS, 
//      METER_SIZE_FOUR_BYTES, 
//      METER_PRECISION_ZERO_DECIMALS, 
//     getter_power, 
//      resetter_power
//    ),
//    ZUNO_METER(
//      ZUNO_METER_TYPE_ELECTRIC, 
//      METER_RESET_ENABLE, 
//      ZUNO_METER_ELECTRIC_SCALE_KWH, 
//      METER_SIZE_FOUR_BYTES, 
//      METER_PRECISION_ZERO_DECIMALS, 
//      getter_energy, 
//      resetter_energy
//    )
//  );

#define ZUNO_CHANNEL_NUMBER_POWER 1
#define ZUNO_CHANNEL_NUMBER_ENERGY 2
     

DWORD getter_power(){
  DEBUG("Getter - Power - Reported value = ");
  DEBUGLN((currentpower);
  return (DWORD)currentpower;
}

DWORD getter_energy(){
  DEBUG("Getter - Energy - Reported value = ");
  DEBUGLN(currentenergy);
  return (DWORD)currentenergy;
}

void resetter_power(byte v) {
  // This function must also reset the energy meter beaseuse the Fibaro HC2 does not call the resetter for the energy meter
  DEBUGLN("Resetter - Power");
  currentpower = 0.0;
  previouspower = 0.0;
  DEBUGLN("Calling resetter_energy"); 
  resetter_energy(0);
}

void resetter_energy(byte v) {
  // This funtion is not used when connected to a Fibaro HC2. Use the resetter_power instead
  DEBUGLN("Resetter - Energy");
  pulsecount = 0; 
  previouspulsecount = 0;
  currentenergy = 0.0;
  previousenergy = 0.0;
  my_meter_data.pulsecount = 0;
  save_meter_data();
}

// For handling of z-uno parameters via controller
word paramvalue;

//ZUNO_SETUP_CFGPARAMETER_HANDLER(config_parameter_changed);


void config_parameter_changed(byte param, word * value) {
  DEBUG("Z-wave parameter received: ");
  DEBUG(param);
  DEBUG(" = ");
  DEBUGLN(*value);
  
  paramvalue = *value;
  
  switch (param) {
  case TICKS_PER_KWH_PARAMETER:
    ppkwh = (unsigned long)*value;
    break;
  case POWER_CHANGE_REPORT_PARAMETER:
    p_changereport = (unsigned long)*value;
    break;
  case POWER_TIME_REPORT_PARAMETER:
    p_timereport = (unsigned long)*value;
    break;
  case ENERGY_CHANGE_REPORT_PARAMETER:
    e_changereport = (unsigned long)*value;
    break;
  case ENERGY_TIME_REPORT_PARAMETER:
    e_timereport = (unsigned long)*value;
    break;
  case ENERGY_SEED_PARAMETER:
    e_seed = (unsigned long)*value;
    pulsecount = e_seed*ppkwh; 
    previouspulsecount = pulsecount;
    currentenergy = (float)pulsecount / ppkwh;
    previousenergy = currentenergy;
    my_meter_data.pulsecount = pulsecount;
    save_meter_data();
    // Reset the value 
    paramvalue = 0;
    
  }
  // Save/report back to the controller
  zunoSaveCFGParam(param, &paramvalue);
   
}

// Set up the interrupt handler on the ZEROX pin (3)
ZUNO_SETUP_ISR_ZEROX(pulse);


/* Interrupt proceure to be called when pulse is recevied
 *  
 */
void pulse (){
  pulsecount++;
  pulsestart = millis();
  pulseactive = true;
  pulsedistance = pulsestart - previouspulse; 
  previouspulse = pulsestart;
  
}


void setup() {

  #ifdef DEBUGMODE
  Serial.begin();
  #endif

  //delay (10000);

    /* Set up values for counters
   * fetch from z-wave parameters 
   */
  zunoLoadCFGParam(TICKS_PER_KWH_PARAMETER, &paramvalue);
  if (paramvalue == 0) {
    ppkwh = TICKS_PER_KWH_DEFAULT;
  } else {
    ppkwh = (unsigned long) paramvalue;
  }
  zunoLoadCFGParam(POWER_CHANGE_REPORT_PARAMETER, &paramvalue);
  if (paramvalue == 0) {
    p_changereport = POWER_CHANGE_REPORT_DEFAULT;
  } else {
    p_changereport = (unsigned long) paramvalue;
  }
  zunoLoadCFGParam(POWER_TIME_REPORT_PARAMETER, &paramvalue);
  if (paramvalue == 0) {
    p_timereport = POWER_TIME_REPORT_DEFAULT;
  } else {
    p_timereport = (unsigned long) paramvalue;
  }
  zunoLoadCFGParam(ENERGY_CHANGE_REPORT_PARAMETER, &paramvalue);
  if (paramvalue == 0) {
    e_changereport = ENERGY_CHANGE_REPORT_DEFAULT;
  } else {
    e_changereport = (unsigned long) paramvalue;
  }
  zunoLoadCFGParam(ENERGY_TIME_REPORT_PARAMETER, &paramvalue);
  if (paramvalue == 0) {
    e_timereport = ENERGY_TIME_REPORT_DEFAULT;
  } else {
    e_timereport = (unsigned long) paramvalue;
  }

  /* Set defualt parameters if not z-uno cfg parameters are used
  ppkwh = TICKS_PER_KWH_DEFAULT;
  p_changereport = POWER_CHANGE_REPORT_DEFAULT;
  p_timereport = POWER_TIME_REPORT_DEFAULT;
  e_changereport = ENERGY_CHANGE_REPORT_DEFAULT;
  e_timereport = ENERGY_TIME_REPORT_DEFAULT;
  */
  ppwh = ppkwh / 1000;
  
  
  // Get saved meter data from EEPROM
  EEPROMreturnval = EEPROM.get(EEPROM_ADDR,  &my_meter_data, sizeof(meter_data));

  // Check data consistency
  if (crc8((byte*)&my_meter_data, sizeof(meter_data) - 1) != my_meter_data.crc8) {
    // Invalid data - reset all
    my_meter_data.pulsecount = 0;
        
    //save_meter_data();
  }
  
  pulsecount = my_meter_data.pulsecount;
  previouspulsecount = my_meter_data.pulsecount;
  currentenergy = (float)pulsecount / ppkwh;
  previousenergy = (float)pulsecount / ppkwh;
  
  
  // Set up output LED which will blnk at same frequency as meter signel
  pinMode(signalpin, OUTPUT);
  
  // Interrupt pin where the pulse counter is attached
  pinMode(pulsepin, INPUT);
  
  previouspowerreport = 0;
  previousenergyreport = 0;


  currentpower = 1.0;
  previouspower = 1.0; 


  DEBUGLN("Starting");

  // initiate the interrupt routine 
  previouspulse = millis();
  zunoExtIntMode(ZUNO_EXT_ZEROX, RISING);

  //Let the power value settle in before starting the main loop
  //delay (5000);

  // Initialize the oled display
  //oled.begin();
  //oled.clrscr();
}

void loop() {
  // put your main code here, to run repeatedly:

  currenttime = millis();

  /* Set the output LED to on if we have received a pulse. Turn it off after PULSE_LENGTD milliseconds
   * 
   */
  if ((pulseactive == true) && ((currenttime - pulsestart) < PULSE_LENGTH)) {
    digitalWrite(signalpin, HIGH);
  } else {
    digitalWrite(signalpin, LOW);
    pulseactive = false;
  }

  
  if (pulsecount > previouspulsecount) {
    // New pulse received
    currentpower = 3600000/(pulsedistance);
    powerdelta = currentpower - previouspower;
    powerincrease = 100.0 * (float)abs(powerdelta) / previouspower;  
/*
    DEBUG("Ver 5.0 - ");
    DEBUGLN(pulsecount);
*/      
    previouspower = currentpower;
    previouspulsecount = pulsecount;

    // Check if power level has increased enough for a report
  
    if (((unsigned long)powerincrease > p_changereport) && (p_changereport > 0)){ 
      // Change in power level higher than report threshold. Send report
      DEBUG("Power increase > ");
      DEBUG((float)p_changereport);
      DEBUGLN("%");
      sendpowerreport = true;  
    
    }
  }
  
  // Check if enough time has passed so we should send a report  
  if (((currenttime - previouspowerreport) > (p_timereport*1000)) && (p_timereport > 0)) {
    // Enough time has elapsed. Send report
    DEBUG("Power report time limit: ");
    DEBUGLN(p_timereport);
    sendpowerreport = true;
  }

  if (sendpowerreport == true) { 
    DEBUG("Sending power report. Power : ");
    DEBUG(currentpower);
    DEBUGLN(" W");
  //  zunoSendReport(ZUNO_CHANNEL_NUMBER_POWER);
    previouspowerreport = currenttime;
    sendpowerreport = false;

  }

  /* Check energy consumption
   *  
   */
  currentenergy = (float)pulsecount / ppkwh;
  energyincrease = currentenergy - previousenergy;
  if ((energyincrease > (float)e_changereport) && (e_changereport > 0)) {
    // Enough energy increase for a report
    DEBUG("Energy change: ");
    DEBUGLN(energyincrease);
    sendenergyreport = true;
  }
  previousenergy = currentenergy;

  // Check if enough time has passed so we should send a report
  if (((currenttime - previousenergyreport) > (e_timereport*1000)) && (e_timereport > 0)) {
    // Enough time has elapsed. Send report
    DEBUG("Energy report time limit: ");
    DEBUGLN(e_timereport);
    
    sendenergyreport = true;
  }

  if (sendenergyreport == true) { 
    DEBUG("Sending energy report. Current energy : ");
    DEBUG2(currentenergy,3);
    DEBUG(" kWh. Last value saved to EEPROM :");
    DEBUG(my_meter_data.pulsecount);
    DEBUG(" pulses (");
    DEBUG2((float)my_meter_data.pulsecount / ppkwh, 3);
    DEBUGLN(" kWh)");
    
  //  zunoSendReport(ZUNO_CHANNEL_NUMBER_ENERGY);
    
    previousenergyreport = currenttime;
    sendenergyreport = false;
  }

  // Check if it is time to save data to EEPROM
  if ((currenttime - lastEEPROMsave) > EEPROM_UPDATE_INTERVAL) {
    my_meter_data.pulsecount = pulsecount;
    save_meter_data();
    lastEEPROMsave = currenttime;
  }

  /*Print status to oled screen
   * 
   */
 //oled.gotoXY(0,0);
 //oled.print(currentpower);
 //oled.print(" W  ");

  
}
